defmodule Clickhouse do
  @moduledoc """
  A Clickhouse client.

  This currently represents an early work in progress.
  """

  @type client :: atom()
  @type statement :: binary()
  @type params :: list()
  @type interface :: module()
  @type format :: module()
  @type option ::
          {:client, client()}
          | {:interface, interface()}
          | {:interface_opts, keyword()}
  @type options :: [option()]

  use Supervisor

  alias Clickhouse.{Error, Format, Interface, Query}

  @opts_schema %{
    client: [type: :atom, required: true, default: :default],
    interface: [type: :module, required: true, default: Clickhouse.Interface.HTTP],
    interface_opts: [type: :list, required: true, default: []]
  }

  ################################
  # Public API
  ################################

  @doc """
  Starts a Clickhouse client.

  ## Options

    * `:client` - The name of the client - defaults to `:default`.
    * `:interface` - The interface to use for queries - defaults to `Clickhouse.Interface.HTTP`.
    * `:interface_opts` - Additional keyword options to pass to the interface.
  """
  @spec start_link(options()) :: Supervisor.on_start()
  def start_link(opts \\ []) do
    with {:ok, opts} <- validate_opts(opts, @opts_schema) do
      Supervisor.start_link(__MODULE__, opts)
    end
  end

  @doc """
  Executes a query using a Clickhouse client.
  """
  @spec query(client(), statement(), params(), keyword()) ::
          {:ok, Clickhouse.Result.t()} | {:error, Clichouse.Error.t()}
  def query(client, statement, params \\ [], opts \\ []) do
    with {:ok, query} <- prepare(client, statement, params, opts),
         {:ok, result} <- execute(query) do
      decode(result)
    end
  end

  ################################
  # Supervisor Callbacks
  ################################

  @impl Supervisor
  def init(opts) do
    init_config(opts)

    children = [
      build_interface(opts)
    ]

    Supervisor.init(children, strategy: :one_for_one)
  end

  ################################
  # Private API
  ################################

  defp validate_opts(opts, schema) do
    case KeywordValidator.validate(opts, schema) do
      {:ok, _} = result -> result
      {:error, reason} -> {:stop, reason}
    end
  end

  defp init_config(opts) do
    client = Keyword.fetch!(opts, :client)
    table_name = table_name(client)
    interface = Keyword.fetch!(opts, :interface)

    :ets.new(table_name, [:named_table, read_concurrency: true])
    :ets.insert(table_name, {:interface, interface})
  end

  defp config(client, key) do
    [{^key, value}] = :ets.lookup(table_name(client), key)
    {:ok, value}
  rescue
    ArgumentError ->
      {:error, Error.new(:not_found, "Clickhouse #{inspect(client)} not found.")}
  end

  defp table_name(client) do
    :"#{__MODULE__}.#{client}"
  end

  defp build_interface(opts) do
    client = Keyword.fetch!(opts, :client)
    interface = Keyword.fetch!(opts, :interface)
    interface_opts = Keyword.fetch!(opts, :interface_opts)
    interface_opts = Keyword.put(interface_opts, :client, client)

    {interface, interface_opts}
  end

  defp prepare(client, statement, params, opts) do
    with {:ok, interface} <- config(client, :interface) do
      query = Query.new(client, interface, statement, params, opts)
      Format.encode(query)
    end
  end

  defp execute(query), do: Interface.execute(query)

  defp decode(result), do: Format.decode(result)
end
