defmodule Clickhouse.Interface.HTTP do
  @moduledoc """
  An interface to interact with a Clickhouse server via HTTP.
  """

  @behaviour Clickhouse.Interface

  use GenServer

  require Logger

  alias Clickhouse.{Error, Format, Query, Result}

  @pool_opts_schema %{
    timeout: [type: :integer, required: true, default: 150_000],
    max_connections: [type: :integer, required: true, default: 50]
  }

  @opts_schema %{
    client: [type: :atom, required: true, default: :default],
    urls: [type: {:list, :binary}, required: true, default: ["http://localhost:8123"]],
    connect_retry: [type: :integer, required: true, default: 3_000],
    pool_opts: [type: {:keyword, @pool_opts_schema}, required: true, default: []]
  }

  @http_opts_schema %{
    checkout_timeout: [type: :timeout, required: false, default: 8_000],
    connect_timeout: [type: :timeout, required: false, default: 8_000],
    recv_timeout: [type: :timeout, required: false, default: 15_000]
  }

  @ping_statement "SELECT 1"

  ################################
  # Clickhouse.Interface Callbacks
  ################################

  @doc """
  Starts the HTTP interface server.
  """
  @impl Clickhouse.Interface
  @spec start_link(keyword()) :: GenServer.on_start()
  def start_link(opts) do
    with {:ok, opts} <- validate_opts(opts, @opts_schema) do
      GenServer.start_link(__MODULE__, opts)
    end
  end

  @doc """
  Executes a query via HTTP.
  """
  @impl Clickhouse.Interface
  @spec execute(Clickhouse.Query.t()) ::
          {:error, Clickhouse.Error.t()} | {:ok, Clickhouse.Result.t()}
  def execute(query) do
    with {:ok, {pool, urls}} <- get_config(query.client),
         {:ok, opts} <- validate_opts(query.opts, @http_opts_schema) do
      statement = Query.get_statement(query)
      do_execute(pool, urls, statement, opts)
    end
  end

  ################################
  # GenServer Callbacks
  ################################

  @impl GenServer
  def init(opts) do
    with {:ok, opts} <- validate_opts(opts, @opts_schema) do
      state = init_state(opts)
      init_config(state)
      init_pool(state)
      {:ok, state, {:continue, :connect}}
    end
  end

  @impl GenServer
  def handle_continue(:connect, state) do
    connect(state)
    {:noreply, state}
  end

  @impl GenServer
  def handle_info(:retry_connect, state) do
    {:noreply, state, {:continue, :connect}}
  end

  ################################
  # Private API
  ################################

  defp validate_opts(opts, schema) do
    case KeywordValidator.validate(opts, schema) do
      {:ok, _} = result -> result
      {:error, reason} -> {:error, Error.new(:argument_error, reason)}
    end
  end

  defp init_state(opts) do
    client = Keyword.get(opts, :client)

    %{
      client: client,
      table_name: table_name(client),
      pool_name: pool_name(client),
      pool_opts: Keyword.get(opts, :pool_opts),
      urls: Keyword.get(opts, :urls),
      connect_retry: Keyword.get(opts, :connect_retry)
    }
  end

  defp init_config(state) do
    config = {state.pool_name, state.urls}
    :ets.new(state.table_name, [:named_table, read_concurrency: true])
    :ets.insert(state.table_name, {:config, config})
  end

  defp get_config(client) do
    [{:config, config}] = :ets.lookup(table_name(client), :config)
    {:ok, config}
  rescue
    ArgumentError ->
      {:error, Error.new(:not_found, "HTTP interface #{inspect(client)} not found.")}
  end

  defp init_pool(state) do
    :hackney_pool.start_pool(state.pool_name, state.pool_opts)
  end

  defp table_name(client) do
    :"#{__MODULE__}.#{client}"
  end

  defp pool_name(client) do
    :"#{__MODULE__}.#{client}.http_pool"
  end

  defp connect(state) do
    case do_execute(state.pool_name, state.urls, @ping_statement, []) do
      {:ok, _} ->
        :ok

      {:error, error} ->
        Logger.error("[Clickhouse] Connection error: #{inspect(error)}")
        Process.send_after(self(), :retry_connect, state.connect_retry)
        :error
    end
  end

  defp do_execute(pool, [url], statement, opts) do
    do_execute(pool, url, statement, opts)
  end

  defp do_execute(pool, urls, statement, opts) when is_list(urls) do
    url = Enum.random(urls)
    do_execute(pool, url, statement, opts)
  end

  defp do_execute(pool, url, statement, opts) when is_binary(url) do
    default_opts = [pool: pool, with_body: true]
    opts = Keyword.merge(opts, default_opts)

    case :hackney.request(:post, url, [], statement, opts) do
      {:ok, 200, headers, body} ->
        build_result(headers, body)

      {:ok, 400, _, body} ->
        {:error, Error.new(:bad_request, body)}

      {:ok, 404, _, body} ->
        {:error, Error.new(:not_found, body)}

      {:ok, code, _, body} when code >= 500 ->
        {:error, Error.new(:server_error, body)}

      {:ok, code, _, _} ->
        {:error, Error.new(:request_error, "HTTP code #{code}")}

      {:error, reason} ->
        {:error, Error.new(reason, nil)}
    end
  end

  defp build_result(headers, body) do
    with {:ok, format} <- parse_format(headers) do
      result = Result.new(body, format)
      {:ok, result}
    end
  end

  defp parse_format([]), do: {:ok, nil}

  defp parse_format([{"X-ClickHouse-Format", format} | _headers]) do
    Format.find(format)
  end

  defp parse_format([_ | headers]) do
    parse_format(headers)
  end
end
