defmodule Clickhouse.Query do
  @moduledoc """
  A Clickhouse query.
  """

  alias Clickhouse.Format

  @type command :: :create | :select | :insert | :alter | :unknown
  @type t :: %__MODULE__{
          client: Clickhouse.client(),
          interface: Clickhouse.interface(),
          format: Clickhouse.format() | nil,
          command: command(),
          statement: Clickhouse.statement(),
          encoded: binary() | nil,
          params: Clickhouse.params(),
          params_count: non_neg_integer()
        }

  @format_regex ~r/(?<= (?:FORMAT|format) )(\w+)/
  @create_regex ~r/\bCREATE\b/i
  @select_regex ~r/\bSELECT\b/i
  @insert_regex ~r/\bINSERT\b/i
  @alter_regex ~r/\bALTER\b/i

  defstruct [
    :client,
    :interface,
    :format,
    :command,
    :statement,
    :encoded,
    :params,
    :params_count,
    :opts
  ]

  ################################
  # Public API
  ################################

  @spec new(
          Clickhouse.client(),
          Clickhouse.interface(),
          Clickhouse.statement(),
          Clickhouse.params(),
          keyword()
        ) ::
          Clickhouse.Query.t()
  def new(client, interface, statement, params \\ [], opts \\ []) do
    %__MODULE__{
      client: client,
      interface: interface,
      format: parse_format(statement),
      command: parse_command(statement),
      statement: statement,
      params: params,
      params_count: parse_params_count(statement),
      opts: opts
    }
  end

  @spec put_encoded(Clickhouse.Query.t(), binary()) :: Clickhouse.Query.t()
  def put_encoded(query, encoded) do
    %{query | encoded: encoded}
  end

  @spec get_statement(Clickhouse.Query.t()) :: binary()
  def get_statement(query) do
    query.encoded || query.statement
  end

  ################################
  # Private API
  ################################

  defp parse_format(statement) do
    case Regex.run(@format_regex, statement, capture: :first) do
      [format] -> get_format(format)
      _ -> nil
    end
  end

  defp get_format(format) do
    case Format.find(format) do
      {:ok, format} -> format
      _ -> nil
    end
  end

  defp parse_command(statement) do
    with {_, false} <- {:create, Regex.match?(@create_regex, statement)},
         {_, false} <- {:select, Regex.match?(@select_regex, statement)},
         {_, false} <- {:insert, Regex.match?(@insert_regex, statement)},
         {_, false} <- {:alter, Regex.match?(@alter_regex, statement)} do
      :unknown
    else
      {command, true} -> command
    end
  end

  defp parse_params_count(statement) do
    statement
    |> String.codepoints()
    |> Enum.count(fn s -> s == "?" end)
  end
end
